@@ -39,6 +39,7 @@ where
     DataToken: Clone + Debug + PartialEq + Eq + Serialize + 'trace,
     for<'de2> DataToken: Deserialize<'de2>,
 {
+    exhausted: bool,
     parent_opid: Opid,
     inner: Rc<RefCell<btree_map::Iter<'trace, Opid, TraceOp<DataToken>>>>,
 }
@@ -52,17 +53,25 @@ where
     type Item = DataToken;
 
     fn next(&mut self) -> Option<Self::Item> {
+        assert!(!self.exhausted);
+
         let (_, trace_op) = advance_ref_iter(self.inner.as_ref())
             .expect("Expected to have an item but found none.");
         assert_eq!(
             self.parent_opid,
             trace_op
                 .parent_opid
-                .expect("Expected an operation with a parent_opid.")
+                .expect("Expected an operation with a parent_opid."),
+            "Expected parent_opid {:?} did not match operation {:#?}",
+            self.parent_opid,
+            trace_op,
         );
 
         match &trace_op.content {
-            TraceOpContent::OutputIteratorExhausted => None,
+            TraceOpContent::OutputIteratorExhausted => {
+                self.exhausted = true;
+                None
+            }
             TraceOpContent::YieldFrom(YieldValue::GetStartingTokens(token)) => Some(token.clone()),
             _ => unreachable!(),
         }
@@ -74,6 +83,7 @@ where
     DataToken: Clone + Debug + PartialEq + Eq + Serialize + 'trace,
     for<'de2> DataToken: Deserialize<'de2>,
 {
+    exhausted: bool,
     parent_opid: Opid,
     data_contexts: Box<dyn Iterator<Item = DataContext<DataToken>> + 'trace>,
     input_batch: VecDeque<DataContext<DataToken>>,
@@ -89,6 +99,7 @@ where
     type Item = (DataContext<DataToken>, FieldValue);
 
     fn next(&mut self) -> Option<Self::Item> {
+        assert!(!self.exhausted);
         let next_op = loop {
             let (_, input_op) = advance_ref_iter(self.inner.as_ref())
                 .expect("Expected to have an item but found none.");
@@ -97,6 +108,9 @@ where
                 input_op
                     .parent_opid
                     .expect("Expected an operation with a parent_opid."),
+                "Expected parent_opid {:?} did not match operation {:#?}",
+                self.parent_opid,
+                input_op,
             );
 
             if let TraceOpContent::AdvanceInputIterator = &input_op.content {
@@ -109,6 +123,9 @@ where
                     input_op
                         .parent_opid
                         .expect("Expected an operation with a parent_opid."),
+                    "Expected parent_opid {:?} did not match operation {:#?}",
+                    self.parent_opid,
+                    input_op,
                 );
 
                 if let TraceOpContent::YieldInto(context) = &input_op.content {
@@ -133,6 +150,7 @@ where
             }
             TraceOpContent::OutputIteratorExhausted => {
                 assert_eq!(None, self.input_batch.pop_front());
+                self.exhausted = true;
                 None
             }
             _ => unreachable!(),
@@ -146,6 +164,7 @@ where
     for<'de2> DataToken: Deserialize<'de2>,
     'trace: 'query,
 {
+    exhausted: bool,
     parent_opid: Opid,
     data_contexts: Box<dyn Iterator<Item = DataContext<DataToken>> + 'query>,
     input_batch: VecDeque<DataContext<DataToken>>,
@@ -162,6 +181,7 @@ where
     type Item = (DataContext<DataToken>, bool);
 
     fn next(&mut self) -> Option<Self::Item> {
+        assert!(!self.exhausted);
         let next_op = loop {
             let (_, input_op) = advance_ref_iter(self.inner.as_ref())
                 .expect("Expected to have an item but found none.");
@@ -170,6 +190,9 @@ where
                 input_op
                     .parent_opid
                     .expect("Expected an operation with a parent_opid."),
+                "Expected parent_opid {:?} did not match operation {:#?}",
+                self.parent_opid,
+                input_op,
             );
 
             if let TraceOpContent::AdvanceInputIterator = &input_op.content {
@@ -182,6 +205,9 @@ where
                     input_op
                         .parent_opid
                         .expect("Expected an operation with a parent_opid."),
+                    "Expected parent_opid {:?} did not match operation {:#?}",
+                    self.parent_opid,
+                    input_op,
                 );
 
                 if let TraceOpContent::YieldInto(context) = &input_op.content {
@@ -207,6 +233,7 @@ where
             }
             TraceOpContent::OutputIteratorExhausted => {
                 assert_eq!(None, self.input_batch.pop_front());
+                self.exhausted = true;
                 None
             }
             _ => unreachable!(),
@@ -220,6 +247,7 @@ where
     for<'de2> DataToken: Deserialize<'de2>,
     'trace: 'query,
 {
+    exhausted: bool,
     parent_opid: Opid,
     data_contexts: Box<dyn Iterator<Item = DataContext<DataToken>> + 'query>,
     input_batch: VecDeque<DataContext<DataToken>>,
@@ -239,6 +267,7 @@ where
     );
 
     fn next(&mut self) -> Option<Self::Item> {
+        assert!(!self.exhausted);
         let next_op = loop {
             let (_, input_op) = advance_ref_iter(self.inner.as_ref())
                 .expect("Expected to have an item but found none.");
@@ -247,6 +276,9 @@ where
                 input_op
                     .parent_opid
                     .expect("Expected an operation with a parent_opid."),
+                "Expected parent_opid {:?} did not match operation {:#?}",
+                self.parent_opid,
+                input_op,
             );
 
             if let TraceOpContent::AdvanceInputIterator = &input_op.content {
@@ -259,6 +291,9 @@ where
                     input_op
                         .parent_opid
                         .expect("Expected an operation with a parent_opid."),
+                    "Expected parent_opid {:?} did not match operation {:#?}",
+                    self.parent_opid,
+                    input_op,
                 );
 
                 if let TraceOpContent::YieldInto(context) = &input_op.content {
@@ -282,6 +317,7 @@ where
                 assert_eq!(trace_context, &input_context);
 
                 let neighbors = Box::new(TraceReaderNeighborIter {
+                    exhausted: false,
                     parent_iterator_opid: next_op.opid,
                     next_index: 0,
                     inner: self.inner.clone(),
@@ -291,6 +327,7 @@ where
             }
             TraceOpContent::OutputIteratorExhausted => {
                 assert_eq!(None, self.input_batch.pop_front());
+                self.exhausted = true;
                 None
             }
             _ => unreachable!(),
@@ -304,6 +341,7 @@ where
     for<'de2> DataToken: Deserialize<'de2>,
     'trace: 'query,
 {
+    exhausted: bool,
     parent_iterator_opid: Opid,
     next_index: usize,
     inner: Rc<RefCell<btree_map::Iter<'trace, Opid, TraceOp<DataToken>>>>,
@@ -321,15 +359,22 @@ where
     fn next(&mut self) -> Option<Self::Item> {
         let (_, trace_op) = advance_ref_iter(self.inner.as_ref())
             .expect("Expected to have an item but found none.");
+        assert!(!self.exhausted);
         assert_eq!(
             self.parent_iterator_opid,
             trace_op
                 .parent_opid
-                .expect("Expected an operation with a parent_opid.")
+                .expect("Expected an operation with a parent_opid."),
+            "Expected parent_opid {:?} did not match operation {:#?}",
+            self.parent_iterator_opid,
+            trace_op,
         );
 
         match &trace_op.content {
-            TraceOpContent::OutputIteratorExhausted => None,
+            TraceOpContent::OutputIteratorExhausted => {
+                self.exhausted = true;
+                None
+            }
             TraceOpContent::YieldFrom(YieldValue::ProjectNeighborsInner(index, token)) => {
                 assert_eq!(self.next_index, *index);
                 self.next_index += 1;
@@ -363,6 +408,7 @@ where
             assert_eq!(vid, vertex_hint);
 
             Box::new(TraceReaderStartingTokensIter {
+                exhausted: false,
                 parent_opid: *root_opid,
                 inner: self.next_op.clone(),
             })
@@ -391,6 +437,7 @@ where
             assert_eq!(*property, field_name);
 
             Box::new(TraceReaderProjectPropertiesIter {
+                exhausted: false,
                 parent_opid: *root_opid,
                 data_contexts,
                 input_batch: Default::default(),
@@ -431,6 +478,7 @@ where
             assert_eq!(eid, &edge_hint);
 
             Box::new(TraceReaderProjectNeighborsIter {
+                exhausted: false,
                 parent_opid: *root_opid,
                 data_contexts,
                 input_batch: Default::default(),
@@ -461,6 +509,7 @@ where
             assert_eq!(*to_type, coerce_to_type_name);
 
             Box::new(TraceReaderCanCoerceIter {
+                exhausted: false,
                 parent_opid: *root_opid,
                 data_contexts,
                 input_batch: Default::default(),
@@ -483,7 +532,9 @@ pub fn assert_interpreted_results<'query, 'trace, DataToken>(
     'trace: 'query,
 {
     let next_op = Rc::new(RefCell::new(trace.ops.iter()));
-    let trace_reader_adapter = Rc::new(RefCell::new(TraceReaderAdapter { next_op }));
+    let trace_reader_adapter = Rc::new(RefCell::new(TraceReaderAdapter {
+        next_op: next_op.clone(),
+    }));
 
     let query: Arc<IndexedQuery> = Arc::new(trace.ir_query.clone().try_into().unwrap());
     let arguments = Arc::new(
@@ -500,13 +551,33 @@ pub fn assert_interpreted_results<'query, 'trace, DataToken>(
         let expected_row = expected_iter.next();
         let trace_row = trace_iter.next();
 
-        if expected_row.is_none() {
+        if let Some(expected_row_content) = expected_row {
+            let trace_expected_row = {
+                let mut next_op_ref = next_op.borrow_mut();
+                let Some((_, trace_op)) = next_op_ref.next() else {
+                    panic!("Reached the end of the trace without producing result {:#?}", trace_row);
+                };
+                let TraceOpContent::ProduceQueryResult(expected_result) = &trace_op.content else {
+                    panic!("Expected the trace to produce a result {:#?} but got another type of operation instead: {:#?}", trace_row, trace_op);
+                };
+                drop(next_op_ref);
+
+                expected_result
+            };
+            assert_eq!(
+                trace_expected_row,
+                expected_row_content,
+                "This trace is self-inconsistent: trace produces row {:#?} but results have row {:#?}",
+                trace_expected_row,
+                expected_row_content,
+            );
+
+            assert_eq!(expected_row, trace_row.as_ref());
+        } else {
             if complete {
                 assert_eq!(None, trace_row);
             }
             return;
-        } else {
-            assert_eq!(expected_row, trace_row.as_ref());
         }
     }
 }
@@ -542,25 +613,28 @@ mod tests {
     }
 
     fn check_filesystem_trace(expected_ir: TestIRQuery, input_data: &str) {
-        if let Ok(test_data) =
-            ron::from_str::<TestInterpreterOutputTrace<FilesystemToken>>(input_data)
-        {
-            assert_eq!(expected_ir.schema_name, "filesystem");
-            assert_eq!(test_data.schema_name, "filesystem");
-            check_trace(expected_ir, test_data);
-        } else {
-            unreachable!()
+        match ron::from_str::<TestInterpreterOutputTrace<FilesystemToken>>(input_data) {
+            Ok(test_data) => {
+                assert_eq!(expected_ir.schema_name, "filesystem");
+                assert_eq!(test_data.schema_name, "filesystem");
+                check_trace(expected_ir, test_data);
+            }
+            Err(e) => {
+                unreachable!("failed to parse trace file: {e}");
+            }
         }
     }
 
     fn check_numbers_trace(expected_ir: TestIRQuery, input_data: &str) {
-        if let Ok(test_data) = ron::from_str::<TestInterpreterOutputTrace<NumbersToken>>(input_data)
-        {
-            assert_eq!(expected_ir.schema_name, "numbers");
-            assert_eq!(test_data.schema_name, "numbers");
-            check_trace(expected_ir, test_data);
-        } else {
-            unreachable!()
+        match ron::from_str::<TestInterpreterOutputTrace<NumbersToken>>(input_data) {
+            Ok(test_data) => {
+                assert_eq!(expected_ir.schema_name, "numbers");
+                assert_eq!(test_data.schema_name, "numbers");
+                check_trace(expected_ir, test_data);
+            }
+            Err(e) => {
+                unreachable!("failed to parse trace file: {e}");
+            }
         }
     }
 
